@@ -6,6 +6,7 @@ from inspect import getfullargspec
 from itertools import chain
 
 from django.conf import settings
+from django.core.cache import cache
 from django.utils.translation import override
 from rest_framework.fields import (
     BooleanField,
@@ -25,7 +26,7 @@ from rest_framework.request import Request
 from rest_framework.serializers import BaseSerializer, Serializer
 
 from .serializers import MockSerializer
-from .typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Type, TypesDict, Union
+from .typing import Any, Callable, Dict, Generator, List, Optional, T, Tuple, Type, TypesDict, Union
 
 
 __all__ = [
@@ -34,13 +35,19 @@ __all__ = [
     "is_serializer_class",
     "serializer_from_callable",
     "inline_serializer",
+    "cache_pipeline_logic",
+    "sentinel",
 ]
 
 
+class sentinel:  # pylint: disable=C0103
+    """Sentinel value."""
+
+
 available_languages: List[str] = [key for (key, value) in settings.LANGUAGES]
 
 
-type_to_serializer_field: Dict[Type, Field] = {
+type_to_serializer_field: Dict[Optional[Type], Type[Field]] = {
     str: CharField,
     int: IntegerField,
     float: FloatField,
@@ -78,7 +85,7 @@ def get_language(request: Request) -> str:
     return "en" if "en" in available_languages else available_languages[0]
 
 
-def translate(item: Union[Callable[..., Any], Request]) -> Union[Generator[Any, Any, None], Callable[..., Any]]:
+def translate(item: Union[Callable[..., T], Request]) -> Union[Generator[Any, Any, None], Callable[..., T]]:
     """Override current language with one from language header or 'lang' parameter.
     Can be used as a context manager or a decorator. If a function is decorated,
     one of the parameters for the function must be a rest_framework.Request object.
@@ -98,7 +105,7 @@ def translate(item: Union[Callable[..., Any], Request]) -> Union[Generator[Any,
                 raise ValueError("No Request-object in function parameters.")
 
             with override(get_language(request)):
-                return item(*args, **kwargs)
+                return item(*args, **kwargs)  # type: ignore
 
         return decorator
 
@@ -110,9 +117,9 @@ def translate(item: Union[Callable[..., Any], Request]) -> Union[Generator[Any,
     return context_manager(item)
 
 
-def _unwrap_types(item: Any) -> Dict:
+def _unwrap_types(item: Any) -> TypesDict:
     """Recurively unwrap types from the given item based on its __annotations__ dict."""
-    annotations = item.__annotations__
+    annotations: TypesDict = item.__annotations__
     for name, annotation in annotations.items():
         if hasattr(annotation, "__annotations__"):
             annotations[name] = _unwrap_types(annotation)
@@ -157,22 +164,24 @@ def parameter_types(func: Callable[..., Any]) -> TypesDict:
     return types
 
 
-def return_types(func: Callable[..., Any]) -> TypesDict:
+def return_types(func: Callable[..., Any]) -> Union[TypesDict, List[TypesDict]]:
     """Get the callables return types"""
     args_spec = getfullargspec(func)
-    types = args_spec.annotations.get("return")
+    types: Union[TypesDict, List[TypesDict]] = args_spec.annotations.get("return")  # type: ignore
+    is_list = hasattr(types, "__args__")
+    types = getattr(types, "__args__", [types])[0]
 
     if hasattr(types, "__annotations__"):
         types = _unwrap_types(types)
 
-    return types
+    return [types] if is_list else types  # type: ignore
 
 
 def get_fields(types: TypesDict) -> Dict[str, Field]:
     """Convert types to serializer fields. TypedDicts and other classes with __annotations__ dicts
     are recursively converted to serializers based on their types.
     """
-    fields = {}
+    fields: Dict[str, Field] = {}
     for name, type_ in types.items():
         if isinstance(type_, dict):
             fields[name] = inline_serializer(name, fields=get_fields(type_))()
@@ -195,9 +204,12 @@ def serializer_from_callable(func: Callable[..., Any], output: bool = False) ->
     In this case, return type should be a TypedDict so that field conversion works.
     """
     types = return_types(func) if output else parameter_types(func)
-    fields: Dict[str, Field] = get_fields(types)
+    is_list = isinstance(types, list)
+    fields: Dict[str, Field] = get_fields(types[0]) if is_list else get_fields(types)  # type: ignore
     serializer_name = snake_case_to_pascal_case(f"{func.__name__}_serializer")
-    return inline_serializer(serializer_name, super_class=MockSerializer, fields=fields)
+    serializer = inline_serializer(serializer_name, super_class=MockSerializer, fields=fields)
+    serializer.many = is_list
+    return serializer  # type: ignore
 
 
 def inline_serializer(
@@ -206,3 +218,27 @@ def inline_serializer(
     fields: Dict[str, Field] = None,
 ) -> Type[BaseSerializer]:
     return type(name, (super_class,), fields or {})  # type: ignore
+
+
+def cache_pipeline_logic(cache_key: str, timeout: int) -> Callable[[Callable[..., T]], Callable[..., T]]:
+    """Cache the result of a pipeline logic function. Calls with different arguments will be saved under
+    different keys, using the given key as a prefix and joining it with a hash of the arguments.
+
+    :param cache_key: Key to save the result under.
+    :param timeout: How long to cache the data in seconds.
+    """
+
+    def decorator(func: Callable[..., T]) -> Callable[..., T]:
+        @wraps(func)
+        def wrapper(*args, **kwargs) -> Any:
+            nonlocal cache_key
+            key = cache_key + str(hash(args + tuple(kwargs.values())))
+            data = cache.get(key, None)
+            if data is None:
+                data = func(*args, **kwargs)
+                cache.set(key, data, timeout)
+            return data
+
+        return wrapper
+
+    return decorator
